module Samplers

import StatsBase: sample

using Random, Distributions, DocStringExtensions, EnsembleKalmanProcesses.ParameterDistributions

export Sampler, FeatureSampler, get_parameter_distribution, get_uniform_shift_bounds, get_rng, sample

"""
$(TYPEDEF)

Wraps the parameter distributions used to sample random features

$(TYPEDFIELDS)
"""
struct Sampler
    "A probability distribution, possibly with constraints"
    parameter_distribution::ParameterDistribution
    "Either `nothing` , or `[lower bound, upper bound]`, which defines a uniform distribution used as a random shift"
    uniform_shift_bounds::Union{AbstractVector, Nothing}
    "A random number generator state"
    rng::AbstractRNG
end

"""
$(TYPEDSIGNATURES)

basic constructor for a `Sampler` 
"""
function FeatureSampler(
    parameter_distribution::ParameterDistribution;
    uniform_shift_bounds::Union{AbstractVector, Nothing} = [0, 2 * pi],
    rng::AbstractRNG = Random.GLOBAL_RNG,
)

    # adds a uniform distribution to the parameter distribution
    if isa(uniform_shift_bounds, AbstractVector)
        if length(uniform_shift_bounds) == 2
            unif_pd = ParameterDistribution(
                Dict(
                    "distribution" => Parameterized(Uniform(uniform_shift_bounds[1], uniform_shift_bounds[2])),
                    "constraint" => no_constraint(),
                    "name" => "uniform",
                ),
            )
            pd = combine_distributions([parameter_distribution, unif_pd])
        end
    else
        pd = parameter_distribution
        uniform_shift_bounds = nothing
    end

    return Sampler(pd, uniform_shift_bounds, rng)

end

"""
$(TYPEDSIGNATURES)

gets the `parameter_distribution` field 
"""
get_parameter_distribution(s::Sampler) = s.parameter_distribution

"""
$(TYPEDSIGNATURES)

gets the `uniform_shift_bounds` field 
"""
get_uniform_shift_bounds(s::Sampler) = s.uniform_shift_bounds

"""
$(TYPEDSIGNATURES)

gets the `rng` field
"""
get_rng(s::Sampler) = s.rng

"""
$(TYPEDSIGNATURES)

samples the distribution within `s`, `n_draws` times using a random number generator `rng`. Can be called without `rng` (defaults to `s.rng`) or `n_draws` (defaults to `1`)
"""
function sample(rng::AbstractRNG, s::Sampler, n_draws::Int)
    pd = get_parameter_distribution(s)
    samp = sample(rng, pd, n_draws)
    constrained_samp = transform_unconstrained_to_constrained(pd, samp)
    #now create a Samples-type distribution from the samples
    s_names = get_name(pd)
    s_slices = batch(pd) # e.g., [1, 2:3, 4:9]
    s_samples = [Samples(constrained_samp[slice, :]) for slice in s_slices]
    s_constraints = [repeat([no_constraint()], size(slice, 1)) for slice in s_slices]

    return combine_distributions([
        ParameterDistribution(ss, sc, sn) for (ss, sc, sn) in zip(s_samples, s_constraints, s_names)
    ])
end

sample(s::Sampler, n_draws::Int) = sample(s.rng, s, n_draws)
sample(rng::AbstractRNG, s::Sampler) = sample(rng, s, 1)
sample(s::Sampler) = sample(s.rng, s, 1)



end # module
